@@ -4,12 +4,16 @@
 #' model fitting objects.
 #'
 #' @param object A fitted model object of appropriate class (e.g.,
-#'   \code{"gbm"}, \code{"lm"}, \code{"randomForest"}, etc.).
+#'   \code{"gbm"}, \code{"lm"}, \code{"randomForest"}, \code{"train"}, etc.).
 #' @param pred.var Character string giving the names of the predictor
 #'   variables of interest. For reasons of computation/interpretation, this
 #'   should include no more than three variables.
-#' @param pred.grid Data frame containing the joint values of the variables
-#'   listed in \code{pred.var}.
+#' @param pred.grid Data frame containing the joint values of interest for the
+#'   variables listed in \code{pred.var}.
+#' @param pred.fun Optional prediction function that requires two arguments:
+#'   \code{object} and \code{newdata}. If specified, then the function must
+#'   return a single prediction or a vector of predictions (i.e., not a matrix
+#'   or data frame). Default is \code{NULL}.
 #' @param grid.resolution Integer giving the number of equally spaced points to
 #'   use (only used for the continuous variables listed in \code{pred.var} when
 #'   \code{pred.grid} is not supplied). If left \code{NULL}, it will default to
@@ -17,8 +21,8 @@
 #'   of the continuous independent variables listed in \code{pred.var}.
 #' @param type Character string specifying the type of supervised learning.
 #'   Current options are \code{"auto"}, \code{"regression"} or
-#'   \code{"classification"}. If \code{type = "auto"} then \code{partial} try
-#'   to extract the necessary information from \code{object}.
+#'   \code{"classification"}. If \code{type = "auto"} then \code{partial} will
+#'   try to extract the necessary information from \code{object}.
 #' @param which.class Integer specifying which column of the matrix of predicted
 #'   probabilities to use as the "focus" class. Default is to use the first
 #'   class. Only used for classification problems (i.e., when
@@ -63,21 +67,43 @@
 #' training data from \code{object}. In these cases an error message is
 #' displayed requesting the user to supply the training data via the
 #' \code{train} argument in the call to \code{partial}. In most cases where
-#' \code{partial} can extract the needed training data from \code{object},
-#' it is taken from the same environment in which \code{partial} was called.
-#' Therefore, it is important to not change the data used to construct
+#' \code{partial} can extract the required training data from \code{object},
+#' it is taken from the same environment in which \code{partial} is called.
+#' Therefore, it is important to not change the training data used to construct
 #' \code{object} before calling \code{partial}. This problem is completely
 #' avoided when the training data are passed to the \code{train} argument in the
 #' call to \code{partial}.
 #'
+#' It is possible to retrieve the last printed \code{"trellis"} object, such as
+#' those produced by \code{plotPartial}, using \code{trellis.last.object()}.
+#'
+#' It is possible for \code{partial} to run much faster if \code{object}
+#' inherits from class \code{"gbm"}. In particular, if \code{object} inherits
+#' from class \code{"gbm"} and \code{pred.grid} is not specified, then
+#' \code{partial} makes an internal call to \code{gbm::plot.gbm} in order to
+#' exploit \code{gbm}'s implementation of the weighted tree traversal method
+#' described in Friedman (2001).
+#'
+#' If the prediction function given to \code{pred.fun} returns a prediction for
+#' each observation in \code{newdata}, then the result will be a PDP for each
+#' observation. These are called individual conditional expectation (ICE)
+#' curves; see Goldstein et al. (2015) and \code{\link[ICEbox]{ice}} for
+#' details.
+#'
 #' @references
 #' J. H. Friedman. Greedy function approximation: A gradient boosting machine.
-#' \emph{Annals of Statistics}, \bold{29}: 1189-1232, 2000.
+#' \emph{Annals of Statistics}, \bold{29}: 1189-1232, 2001.
+#'
+#' Goldstein, A., Kapelner, A., Bleich, J., and Pitkin, E., Peeking Inside the
+#' Black Box: Visualizing Statistical Learning With Plots of Individual
+#' Conditional Expectation. (2014) \emph{Journal of Computational and Graphical
+#' Statistics}, \bold{24}(1): 44-65, 2015.
 #'
 #' @rdname partial
 #' @export
 #' @examples
 #' \dontrun{
+#'
 #' #
 #' # Regression example (requires randomForest package to run)
 #' #
@@ -97,7 +123,7 @@
 #'
 #' # The partial function allows for multiple predictors
 #' partial(boston.rf, pred.var = c("lstat", "rm"), grid.resolution = 40,
-#'         plot = TRUE, chull = TRUE, .progress = "text")
+#'         plot = TRUE, chull = TRUE, progress = "text")
 #'
 #' # The plotPartial function offers more flexible plotting
 #' pd <- partial(boston.rf, pred.var = c("lstat", "rm"), grid.resolution = 40)
@@ -114,9 +140,17 @@
 #' set.seed(102)  # for reproducibility
 #' pima.rf <- randomForest(diabetes ~ ., data = pima, na.action = na.omit)
 #'
-#' # Partial dependence of glucose on diabetes test result (neg/pos)
+#' # Partial dependence of diabetes test result (neg/pos) on glucose
 #' partial(pima.rf, pred.var = c("glucose", "age"), plot = TRUE, chull = TRUE,
-#'         .progress = "text")
+#'         progress = "text")
+#'
+#' # Partial dependence of positive diabetes test result on glucose, plotted on
+#' # the probability scale, rather than the centered logit
+#' pfun <- function(object, newdata) {
+#'   mean(predict(object, newdata, type = "prob")[, "pos"], ne.rm = TRUE)
+#' }
+#' partial(pima.rf, pred.var = "glucose", pred.fun = pfun,
+#'         plot = TRUE, chull = TRUE, progress = "text")
 #'
 #' #
 #' # Interface with caret (requires caret package to run)
@@ -137,6 +171,7 @@
 #'
 #' # Partial dependence of glucose on diabetes test result (neg/pos)
 #' partial(pima.svm, pred.var = "glucose", plot = TRUE, rug = TRUE)
+#'
 #' }
 partial <- function(object, ...) {
   UseMethod("partial")
@@ -145,92 +180,136 @@ partial <- function(object, ...) {
 
 #' @rdname partial
 #' @export
-partial.default <- function(object, pred.var, pred.grid, grid.resolution = NULL,
+partial.default <- function(object, pred.var, pred.grid, pred.fun = NULL,
+                            grid.resolution = NULL,
                             type = c("auto", "regression", "classification"),
                             which.class = 1L, plot = FALSE,
                             smooth = FALSE, rug = FALSE, chull = FALSE, train,
                             check.class = TRUE, progress = "none",
                             parallel = FALSE, paropts = NULL, ...) {
 
-  # Error message to display when training data cannot be extracted form object
-  mssg <- paste0("The training data could not be extracted from ",
-                 deparse(substitute(object)), ". Please supply the raw ",
-                 "training data using the `train` argument in the call to ,",
-                 "`partial`.")
+  # Construct partial dependence data
+  if (inherits(object, "gbm") && missing(pred.grid) && is.null(pred.fun)) {
 
-  # Data frame
-  if (missing(train)) {
-    if (inherits(object, "BinaryTree") || inherits(object, "RandomForest")) {
-      train <- object@data@get("input")
-    } else if (inherits(object, "train")) {
-      train <- object$trainingData
-      train$.outcome <- NULL  # remove .outcome column
-    } else if (isS4(object)) {
-      stop(mssg)
-    } else {
-      if (is.null(object$call$data)) {
-        stop(mssg)
+    # FIXME: which.class gets ignored by gbm::plot.gbm
+
+    # If not supplied, try to extract training data from object when plot = TRUE
+    if ((plot || chull) && missing(train)) {
+      train <- getTrainingData(object)
+    }
+
+    # If grid.resolution is NULL, use the same default as continuous.resolution
+    # in gbm::plot.gbm.
+    if (is.null(grid.resolution)) {
+      grid.resolution <- 100
+    }
+
+    # Call gbm::plot.gbm directly; this is MUCH FASTER!
+    pd.df <- gbm::plot.gbm(object, i.var = pred.var,
+                           continuous.resolution = grid.resolution,
+                           return.grid = TRUE, ...)
+
+    # Restrict to convex hull of first two predictors, if requested
+    if (chull) {
+      pd.df <- if (length(pred.var) >= 2 && is.numeric(pd.df[[1L]]) &&
+          is.numeric(pd.df[[2L]])) {
+        X <- stats::na.omit(data.matrix(train[pred.var[1L:2L]]))
+        Y <- stats::na.omit(data.matrix(pd.df[, -ncol(pd.df),
+                                              drop = FALSE][1L:2L]))
+        hpts <- grDevices::chull(X)
+        hpts <- c(hpts, hpts[1L])
+        keep <- mgcv::in.out(X[hpts, ], Y)
+        pd.df[keep, ]
       } else {
-        train <- eval(object$call$data)
+        pd.df
       }
     }
-  }
 
-  # Throw error message if predictor names not found in training data
-  if (!all(pred.var %in% names(train))) {
-    stop(paste(paste(pred.var[!(pred.var %in% names(train))], collapse = ", "),
-               "not found in the training data."))
-  }
-
-  # Throw an error message of one of the predictors is labelled "y"
-  if ("y" %in% pred.var) {
-    stop("\"y\" cannot be a predictor name.")
-  }
-
-  # NOTE: It is worth considering the approach taken by the plotmo package,
-  # which now has the option to compute PDPs. plotmo seems to make one large
-  # pred.grid which increases memory usage, but decreases the number of calls to
-  # stats::predict.
-
-  # Predictor values of interest
-  if (missing(pred.grid)) {
-    pred.grid <- predGrid(object, pred.var = pred.var, train = train,
-                          grid.resolution = grid.resolution)
-  }
-
-  # Make sure each column has the correct class, factor levels, etc.
-  if (check.class) {
-    pred.grid <- copyClasses(pred.grid, train)
-  }
-
-  # Restrict grid to covext hull of first two columns
-  if (chull) {
-    pred.grid <- trainCHull(pred.var, pred.grid = pred.grid, train = train)
-  }
-
-  # Determine the type of supervised learning used
-  type <- match.arg(type)
-  if (type == "auto") {
-    type <- superType(object)
-  }
-
-  # Calculate partial dependence values
-  if (type == "regression") {
-    pd.df <- pdRegression(object, pred.var = pred.var, pred.grid = pred.grid,
-                          train = train, progress = progress,
-                          parallel = parallel, paropts = paropts, ...)
-  } else if (type == "classification") {
-    pd.df <- pdClassification(object, pred.var = pred.var,
-                              pred.grid = pred.grid, which.class = which.class,
-                              train = train, progress = progress,
-                              parallel = parallel, paropts = paropts, ...)
   } else {
-    stop(paste("Partial dependence values are currently only available",
-               "for classification and regression problems."))
+
+    # Match pred.var function if not NULL
+    if (!is.null(pred.fun)) {
+      pred.fun <- match.fun(pred.fun)
+      if (!identical(names(formals(pred.fun)), c("object", "newdata"))) {
+        stop(paste0("pred.fun requires a function with only two arguments: ",
+                    "object, and newdata."))
+      }
+    }
+
+    # If not supplied, try to extract training data from object
+    if (missing(train)) {
+      train <- getTrainingData(object)
+    }
+
+    # Allow column position specification
+    if (is.numeric(pred.var)) {
+      pred.var <- names(train)[pred.var]
+    }
+
+    # Throw error message if predictor names not found in training data
+    if (!all(pred.var %in% names(train))) {
+      stop(paste(paste(pred.var[!(pred.var %in% names(train))], collapse = ", "),
+                 "not found in the training data."))
+    }
+
+    # Throw an error message of one of the predictors is labelled "y"
+    if ("y" %in% pred.var) {
+      stop("\"y\" cannot be a predictor name.")
+    }
+
+    # Predictor values of interest
+    if (missing(pred.grid)) {
+      pred.grid <- predGrid(object, pred.var = pred.var, train = train,
+                            grid.resolution = grid.resolution)
+    }
+
+    # Make sure each column has the correct class, factor levels, etc.
+    if (check.class) {
+      pred.grid <- copyClasses(pred.grid, train)
+    }
+
+    # Restrict grid to covext hull of first two columns
+    if (chull) {
+      pred.grid <- trainCHull(pred.var, pred.grid = pred.grid, train = train)
+    }
+
+    # Determine the type of supervised learning used
+    type <- match.arg(type)
+    if (type == "auto") {
+      type <- superType(object)
+    }
+
+    # Calculate partial dependence values
+    if (type == "regression") {
+      pd.df <- pdRegression(object, pred.var = pred.var, pred.grid = pred.grid,
+                            pred.fun = pred.fun, train = train,
+                            progress = progress, parallel = parallel,
+                            paropts = paropts, ...)
+    } else if (type == "classification") {
+      pd.df <- pdClassification(object, pred.var = pred.var,
+                                pred.grid = pred.grid, pred.fun = pred.fun,
+                                which.class = which.class, train = train,
+                                progress = progress, parallel = parallel,
+                                paropts = paropts, ...)
+    } else {
+      stop(paste("Partial dependence values are currently only available",
+                 "for classification and regression problems."))
+    }
+
   }
 
-  # Create data frame of partial dependence values
-  names(pd.df) <- c(pred.var, "yhat")
+  # Construct an appropriate data frame
+  if (any(grepl("^yhat\\.", names(pd.df)))) {  # assume ICE curves
+    # Convert from wide to long format
+    pd.df <- stats::reshape(pd.df, varying = (length(pred.var) + 1):ncol(pd.df),
+                            direction = "long")
+    pd.df$id <- NULL  # remove id column
+    pd.df <- pd.df[, c(pred.var, "yhat", "time")]  # rearrange columns
+    names(pd.df)[ncol(pd.df)] <- "yhat.id"  # rename "time" column
+  } else {
+    names(pd.df) <- c(pred.var, "yhat")
+  }
+  rownames(pd.df) <- NULL  # remove row names
   class(pd.df) <- c("data.frame", "partial")
 
   # Plot partial dependence function (if requested)
